package szTools.console;

import java.util.Hashtable;
import java.io.PrintStream;
import java.io.ByteArrayOutputStream;
import java.io.PrintWriter;

public final class TextConsole implements Runnable,Console {
  /** 
   * Contains info about registered writers. 
   * @see #addWriter
   * @see #messageFrom
   * @see #exception
   */
  private Hashtable writers;

  /** 
   * Currently served writer.
   * @see #addWriter
   * @see #messageFrom
   * @see #exception
   */
  private Object currentWriter;

  /** 
   * Waits for messages from writers.
   * @see #run
   */ 
  private Thread listener;

  /** 
   * Message to be printed. 
   * @see #messageFrom
   */
  private String message;

  /** 
   * Exception to be shown. 
   * @see #exception
   */
  private Exception exception;

  /** 
   * Stops the listener thread. 
   * @see #listener
   */
  private boolean stop;

  /** 
   * Contains handle of last created instance. 
   * @see #getInstance
   */
  private static TextConsole instance=null;

  /** 
   * @return  handle of last created instance
   * @see     #instance
   */
  public Console getInstance() {
    return instance;
  }

  public TextConsole() {
    this( System.err );
  }

  PrintStream stream;

  /** Constructs a console. */
  public TextConsole( PrintStream str ) {
    stream = str;

    // utworzenie listy pisarzy (na razie pustej)
    writers = new Hashtable();

    // uruchamiamy watek nasluchujacy
    listener = new Thread( this );
    listener.setPriority( Thread.MIN_PRIORITY );  // ten watek zbyt wazny nie jest
    listener.start();

    // WAZNE!!!
    instance = this;
    stream.println( "Text Console v. 0.4 (SZ) ready." );
  }

  /** 
   * The body of the listener thread.
   * @see listener
   */
  public synchronized void run() {
    WriterData writerData;
    ByteArrayOutputStream ostr;
    PrintWriter wr;

    stop = false;
    while( !stop ) {
      while( message==null && exception==null ) {
        try{
          wait();
        }
        catch( InterruptedException e ) {}
      }
      
      if( exception == null ) {        // standardowa wiadomosc
        if( currentWriter != null ) {
          writerData = getWriterData( currentWriter );
	  if( writerData != null ) {
	    if( writerData.isEnabled() ) {
	      stream.println( "Message from " + writerData.getName() + ": " + message );
	    }
	  }
        }
        else {
	  stream.println( "Message: " +message );
	}
      }
      else {           // wyjatek - nie sprawdzamy maskowania
	wr = new PrintWriter( ostr=new ByteArrayOutputStream() );
	exception.printStackTrace( wr );
	System.err.println( "Exception!" );
	exception.printStackTrace();

      	if( currentWriter != null ) {
	  writerData = getWriterData( currentWriter );
	  stream.println( "Exception (in "+writerData.getName()+" code): " +exception+ '\n' + ostr.toString() );
	}
        else {
	  stream.println( "Exception: " +exception+ '\n' +ostr.toString() );
	}
      }
      exception = null;
      currentWriter = null;
      message = null;
    }
  }

  /** 
   * Writes a message with no source specified. 
   * @param txt message text
   * @see #messageFrom
   */
  public void trace( String txt ) {
    if( instance != null ) {
      instance.currentWriter = null;
      instance.message = txt;
      synchronized(instance) {
        instance.notify();
      }
    }
  }

  /** 
   * Registers a new writer.
   * @param  writer        object to be registered
   * @param  name          object signature (to be shown with messages)
   * @see    #messageFrom
   * @see    #exception
   */
  public void addWriter( Object writer, String name ) {
    writers.put( writer,new WriterData( name ) );
  }

  /** 
   * Prints out a message.
   * @param  writer       message sender
   * @param  txt          message text
   * @see    #addWriter
   * @see    #exception 
   */
  public void messageFrom( Object writer, String txt ) {
    currentWriter = writer;
    message = txt;
    synchronized(this) {
      notify();
    }    
  }

  /** 
   * Reports an exception. 
   * @param  writer       object in which code the exception occured
   * @param  exception    exception to be printed out
   * @see    #addWriter
   * @see    #messageFrom
   */
  public void exception( Object writer, Exception exception ) {
    currentWriter = writer;
    this.exception = exception;
    synchronized(this) {
      notify();
    }
  }

  /**
   * Encapsulates data describing writer's state.
   */
  private class WriterData {
    private String name;
    private boolean enabled;

    /**
     * Construct new record.
     * @param name        writer's signature
     * @see   GraphicalConsole#addWriter
     */
    WriterData( String name ) {
      this.name = new String( name );
      this.enabled = true;
    }

    /**
     * Enables / disables printing out messages sent by the writer.
     * @param en       new state of the 'enabled' flag
     * @see   #enabled
     * @see   #isEnabled
     */
    void setEnabled( boolean en ) {
      this.enabled = en;
    }

    /** 
     * Checks the 'enabled' flag.
     * @return <CODE>true</CODE> if writer is enabled, 
     * <CODE>false</CODE> otherwise
     * @see #enabled
     * @see #setEnabled
     */ 
    boolean isEnabled() {
      return enabled;
    }

    /**
     * Returns the writer's signature.
     */
    String getName() {
      return name;
    }
  }


  /**
   * Returns handle to writer's data record.
   * @param  writer handle of the object to describe
   * @see    #writers
   */
  private WriterData getWriterData( Object writer ) {
    return (WriterData)(writers.get(writer));
  }



  /** Test program. */
  public static void main( String[] args ) {
    TextConsole console = new TextConsole();
  }

  public void display() {
  }

} // finish class GraphicalConsole











